// Copyright (c) 2011 Sundog Software, LLC. All rights reserved worldwide.

#include "nvsg/GeoNode.h"
#include "nvsg/Camera.h"
#include "nvsg/DirectedLight.h"
#include "nvsg/BlendAttribute.h"
#include "nvsg/Material.h"
#include "nvmath/Spherent.h"
#include "SilverLiningManager.h"

using namespace nvsg;
using namespace nvutil;
using namespace nvmath;
using namespace SilverLining;

typedef struct SilverLiningData_S
{
    SilverLining::Atmosphere *atm;
    CameraSharedPtr camera;
    DirectedLightHandle *light;
} SilverLiningData;

Atmosphere *SilverLiningManager::AttachToRoot(GroupSharedPtr rootNode, CameraSharedPtr camera,
                                              DirectedLightHandle *light)
{
    // Create an atmosphere object:
    Atmosphere *atm = new Atmosphere("Your user name", "Your license key");

    const char *slPath = getenv("SILVERLINING_PATH");
    std::string resourcePath(slPath);
    resourcePath += "\\resources\\";

    int result = atm->Initialize(Atmosphere::OPENGL, resourcePath.c_str(), true, NULL);
    if (result != Atmosphere::E_NOERROR)
    {
        return NULL;
    }

    // Populate data we want to have access to
    SilverLiningData *slData = new SilverLiningData;
    slData->atm = atm;
    slData->camera = camera;
    slData->light = light;

    // Placeholder for a custom DrawCallback. Due to a bug in SceniX 7.0, DrawCallbacks
    // on custom drawable objects don't work - they only work on Primitive subclasses.
    // So to work around the issue, we need to construct a dummy Primitive with some
    // vertices that don't actually get used.
    VertexAttributeSetSharedPtr vertexAttributeSet = VertexAttributeSet::create();  
    {    
        std::vector<Vec3f> vertices;    
        vertices.push_back(Vec3f(-0.5f, -0.5f, 0.0f));    
        vertices.push_back(Vec3f( 0.5f,  0.0f, 0.0f));    
        vertices.push_back(Vec3f( 0.0f,  0.5f, 0.0f));    
        VertexAttributeSetWriteLock vas( vertexAttributeSet );    
        vas->setVertices( &vertices[0], 3 );  
    }

    // Attach the sky:
    SilverLiningSkySharedPtr hdlSky = SilverLiningSkyHandle::create();
    {
        PrimitiveWriteLock theSky(hdlSky);
        theSky->setPrimitiveType(PRIMITIVE_POINTS);
        theSky->setVertexAttributeSet(vertexAttributeSet);
        theSky->setUserData(slData);
        theSky->setDrawCallback(new SilverLiningSkyDraw(atm));
        theSky->addHints(Object::NVSG_HINT_ALWAYS_VISIBLE);
    }

    // Assertions fail if we don't have at least one attribute, so
    // we'll create a dummy one.
    MaterialSharedPtr hdlMat = Material::create();    
    StateSetSharedPtr statePtr = StateSet::create();
    {
        StateSetWriteLock sswl(statePtr);
        sswl->addAttribute(hdlMat);
    }

    // Create the GeoNode containing our primitive
    GeoNodeWeakPtr hdlSkyNode = GeoNodeHandle::create();
    {
        GeoNodeWriteLock theSkyNode(hdlSkyNode);
        theSkyNode->addDrawable(statePtr, hdlSky);
        theSkyNode->addHints(Object::NVSG_HINT_ALWAYS_VISIBLE);
    }

    // Add it to the group.
    GroupWriteLock groupWriteLock(rootNode);
    groupWriteLock->addChild(hdlSkyNode);

    // Attach the clouds:
    SilverLiningCloudsSharedPtr hdlClouds = SilverLiningCloudsHandle::create();
    {
        PrimitiveWriteLock theClouds(hdlClouds);
        theClouds->setPrimitiveType(PRIMITIVE_POINTS);
        theClouds->setVertexAttributeSet(vertexAttributeSet);
        theClouds->setUserData(slData);
        theClouds->setDrawCallback(new SilverLiningCloudsDraw(atm));
    }

    GeoNodeWeakPtr hdlCloudNode = GeoNodeHandle::create();

    GeoNodeWriteLock gnwl(hdlCloudNode);
    gnwl->addDrawable(statePtr, hdlClouds);
	gnwl->setName( "SilverLiningCloudsGeometry" );

    // Add it to the group as well.
    groupWriteLock->addChild(hdlCloudNode);

    // Return our Atmosphere object so you can configure it
    return atm;
}

void SilverLiningSkyDraw::updateLight(DirectedLightHandle *lightHdl)
{
    if (!lightHdl || !atmosphere) return;

    float rd, gd, bd; // directional light
    float ra, ga, ba; // ambient light
    float x, y, z;

    atmosphere->GetSunOrMoonColor(&rd, &gd, &bd);
    atmosphere->GetAmbientColor(&ra, &ga, &ba);
    atmosphere->GetSunOrMoonPosition(&x, &y, &z);

    WritableObject<DirectedLight> light(lightHdl);
    light->setEnabled(true);
    light->setAmbientColor(nvmath::Vec3f(ra, ga, ba));
    light->setDiffuseColor(nvmath::Vec3f(rd, gd, bd));
    light->setSpecularColor(nvmath::Vec3f(rd, gd, bd));
    light->setDirection(nvmath::Vec3f(-x, -y, -z));
}

void SilverLiningSkyDraw::extractMatrices(CameraSharedPtr camHdl)
{
    if (!camHdl || !atmosphere) return;

    CameraReadLock cam(camHdl);

    nvmath::Mat44f viewMat = cam->getWorldToViewMatrix();
    const float *fViewMat = viewMat.getPtr();
    double dCamMat[16];
    for (int i = 0; i < 16 ;i++)
    {
        dCamMat[i] = fViewMat[i];
    }
    atmosphere->SetCameraMatrix(dCamMat);   

    nvmath::Mat44f projMat = cam->getProjection();
    const float *fProjMat = projMat.getPtr();
    double dProjMat[16];
    for (int j = 0; j < 16; j++)
    {
        dProjMat[j] = fProjMat[j];
    }
    atmosphere->SetProjectionMatrix(dProjMat);
}

bool SilverLiningSkyDraw::draw(const Drawable *drawable)
{
    SilverLiningData *slData = (SilverLiningData *)drawable->getUserData();
    if (slData)
    {
        // Extract the view and projection matrices and pass them to SilverLining:
        extractMatrices(slData->camera);

        // Draw the sky box:
        atmosphere->DrawSky(true);

        // Sync up the directional light:
        updateLight(slData->light);
    }

    // Tell SceniX we handled the drawing ourselves:
    return true;
}

bool SilverLiningCloudsDraw::draw(const Drawable *drawable)
{
    // Draw the clouds:
    atmosphere->DrawObjects();
    
    // Tell SceniX we handled the drawing ourselves:
    return true;
}

SilverLiningSky::~SilverLiningSky()
{
    SilverLiningData *slData = (SilverLiningData *)getUserData();
    if (slData)
    {
        delete slData->atm;
        delete slData;
    }
}

nvmath::Sphere3f  SilverLiningSky::calculateBoundingSphere () const
{
    // Compute the size and position of the sky box.

    SilverLiningData *slData = (SilverLiningData *)getUserData();
    float skyBoxRadius = 2000.0f;
    nvmath::Vec3f camPos(0, 0, 0);

    if (slData)
    {
        double boxSize = slData->atm->GetConfigOptionDouble("sky-box-size");
        skyBoxRadius = (float)boxSize * 1.42f;

        CameraSharedPtr camHandle = slData->camera;
        CameraReadLock cam(camHandle);
        camPos = cam->getPosition();
    }

    nvmath::Sphere3f s;
    s.setCenter(camPos);
    s.setRadius(skyBoxRadius);

    return s;
}

nvmath::Box3f  SilverLiningSky::calculateBoundingBox () const
{
    // Compute the size and position of the sky box.

    SilverLiningData *slData = (SilverLiningData *)getUserData();
    float skyBoxDim = 2000.0f;
    nvmath::Vec3f camPos(0, 0, 0);

    if (slData)
    {
        skyBoxDim = (float)slData->atm->GetConfigOptionDouble("sky-box-size");

        CameraSharedPtr camHandle = slData->camera;
        CameraReadLock cam(camHandle);
        camPos = cam->getPosition();
    }

    nvmath::Box3f box(nvmath::Vec3f(-skyBoxDim, -skyBoxDim, -skyBoxDim) + camPos,
        nvmath::Vec3f(skyBoxDim, skyBoxDim, skyBoxDim) + camPos);

    return box;
   
}

nvmath::Sphere3f  SilverLiningClouds::calculateBoundingSphere () const
{
    // Compute the size and position of all cloud layers.
    // Since we're only using camera-centered "infinite" cloud layers, this is 
    // fairly simple. If you had non-infinite cloud layers, you'd need to 
    // intersect the bounds of each layer here taking their positions into account.

    SilverLiningData *slData = (SilverLiningData *)getUserData();

    nvmath::Vec3f cloudsCenter(0, 0, 0);
    float cloudsRadius = 100000;

    if (slData)
    {
        CameraSharedPtr camHandle = slData->camera;
        CameraReadLock cam(camHandle);
        cloudsCenter = cam->getPosition();
        cloudsCenter[1] = 0; // Note, we're assuming the default y-is-up convention here! Change this otherwise.

        Atmosphere *atm = slData->atm;
        SL_MAP(int, CloudLayer*)& cloudLayers = atm->GetConditions()->GetCloudLayers();
        SL_MAP(int, CloudLayer*)::iterator it;

        float maxRadius = 0;

        for (it = cloudLayers.begin(); it != cloudLayers.end(); it++)
        {
            int cloudHandle = it->first;
            CloudLayer *layer = it->second;

            double length = layer->GetBaseLength();
            double width = layer->GetBaseWidth();
            double thickness = layer->GetThickness();
            double altitude = layer->GetBaseAltitude();
        
            double yr = thickness + altitude;
            double xr = width * 0.5;
            double zr = length * 0.5;
            double radius = sqrt(xr*xr + yr*yr + zr*zr);
            
            if (radius > maxRadius)
            {
                maxRadius = (float)radius;
            }
        }

        cloudsRadius = maxRadius;
    }

    nvmath::Sphere3f s;
    s.setCenter(cloudsCenter);
    s.setRadius(cloudsRadius);

    return s;
}

nvmath::Box3f  SilverLiningClouds::calculateBoundingBox () const
{
    // Compute the size and position of all cloud layers.
    // Since we're only using camera-centered "infinite" cloud layers, this is 
    // fairly simple. If you had non-infinite cloud layers, you'd need to 
    // intersect the bounds of each layer here taking their positions into account.

    SilverLiningData *slData = (SilverLiningData *)getUserData();

    nvmath::Vec3f cloudsCenter(0, 0, 0);
    float cloudsRadius = 100000;

    nvmath::Box3f box;

    if (slData)
    {
        CameraSharedPtr camHandle = slData->camera;
        CameraReadLock cam(camHandle);
        cloudsCenter = cam->getPosition();
        cloudsCenter[1] = 0; // Note, we're assuming the default y-is-up convention here! Change this otherwise.

        Atmosphere *atm = slData->atm;
        SL_MAP(int, CloudLayer*)& cloudLayers = atm->GetConditions()->GetCloudLayers();
        SL_MAP(int, CloudLayer*)::iterator it;

        float maxRadius = 0;

        for (it = cloudLayers.begin(); it != cloudLayers.end(); it++)
        {
            int cloudHandle = it->first;
            CloudLayer *layer = it->second;

            double length = layer->GetBaseLength();
            double width = layer->GetBaseWidth();
            double thickness = layer->GetThickness();
            double altitude = layer->GetBaseAltitude();
        
            double yr = thickness + altitude;
            double xr = width * 0.5;
            double zr = length * 0.5;

            box.update(nvmath::Vec3f((float)xr, (float)yr, (float)zr) + cloudsCenter);
        }
    }

    return box;
}
