// Copyright NVIDIA Corporation 2011
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES

// Copyright Sundog Software 2011
// This modified GLUTMinimal main.cpp integrates in a SilverLiningManager, and configures
// some clouds for it. Our modifications are preceded by comments beginning with SUNDOG

#include <GL/freeglut.h>
#include <nvsg/nvsg.h>
#include <nvgl/SceneRendererGL2.h>
#include <nvgl/RenderTargetGLFB.h>
#include <nvsg/CoreTypes.h>
#include <nvsg/Node.h>
#include <nvsg/DirectedLight.h>
#include <nvsg/Group.h>
#include <nvsg/FrustumCamera.h>
#include <nvsg/ViewState.h>
#include "SceneFunctions.h"
#include "SimpleScene.h"

// SUNDOG:
#include "SilverLiningManager.h"

using namespace nvutil;
using namespace nvsg;
using namespace nvmath;

nvgl::SmartRenderTargetGL   rtgl;   // RenderTarget for SceneRenderer
nvgl::SmartSceneRendererGL2 srgl2;  // OpenGL2 based renderer

// SUNDOG:
SilverLiningManager *silverLining = 0;

// NOTE: this sample program is described in-depth in the documentation,
// in the "Getting Started" part of the SceniX Programming Guide.


// the sample's display function, called when the window needs to be redisplayed
void display()
{
  // render the scene
  srgl2->render();

  // tell GLUT to display the new rendered image
  glutSwapBuffers();

  glutPostRedisplay();
}

// the sample's reshape function, called when the window is reshaped
void reshape( GLint width, GLint height )
{
  // pass the new window size to the render target
  rtgl->setSize( width, height );
}


// SUNDOG: Function to configure the atmosphere and add in some clouds.
void setupAtmosphericConditions(SilverLining::Atmosphere *atm)
{
    if (!atm) return;

    // Set the visibility so clouds fade out in the distance
    atm->GetConditions()->SetVisibility(30000.0);

    // Set the simulated location to somewhere near Seattle
    SilverLining::Location loc;
    loc.SetAltitude(0);
    loc.SetLatitude(47);
    loc.SetLongitude(-122);
    atm->GetConditions()->SetLocation(loc);

    // Set the time to midsummer - it's important that the time zone
    // specified matches the position above!
    SilverLining::LocalTime tm;
    tm.SetTimeZone(PST);
    tm.SetYear(2010);
    tm.SetMonth(7);
    tm.SetDay(10);
    tm.SetHour(17);
    tm.SetMinutes(0);
    tm.SetSeconds(0);
    atm->GetConditions()->SetTime(tm);

    // Add some cirrus clouds
    SilverLining::CloudLayer *cirrusCloudLayer = SilverLining::CloudLayerFactory::Create(CIRRUS_FIBRATUS);
    cirrusCloudLayer->SetIsInfinite(true);
    cirrusCloudLayer->SetBaseAltitude(8000);
    cirrusCloudLayer->SetThickness(0);
    cirrusCloudLayer->SetBaseLength(100000);
    cirrusCloudLayer->SetBaseWidth(100000);
    cirrusCloudLayer->SetLayerPosition(0, 0);
    cirrusCloudLayer->SeedClouds(*atm);

    atm->GetConditions()->AddCloudLayer(cirrusCloudLayer);

    // And some cumulus congestus
    SilverLining::CloudLayer *cumulusCongestusLayer;
    cumulusCongestusLayer = SilverLining::CloudLayerFactory::Create(CUMULUS_CONGESTUS);
    cumulusCongestusLayer->SetIsInfinite(true);
    cumulusCongestusLayer->SetBaseAltitude(1500);
    cumulusCongestusLayer->SetThickness(100);
    cumulusCongestusLayer->SetBaseLength(30000);
    cumulusCongestusLayer->SetBaseWidth(30000);
    cumulusCongestusLayer->SetDensity(0.4);
    cumulusCongestusLayer->SetLayerPosition(0, 0);
    cumulusCongestusLayer->SeedClouds(*atm);

    atm->GetConditions()->AddCloudLayer(cumulusCongestusLayer);
}

// the sample's main function
// generate a scene, set up the renderer and enter the GLUT main loop
int sampleMain()
{
  // create a SimpleScene object, containing a SceniX scene
  SimpleScene scene;

  // create a ViewState object, containing view specific state like a camera and a scene
  nvsg::ViewStateSharedPtr viewState = nvsg::ViewState::create();

  // pass the scene contained in the SimpleScene object to the ViewState object
  nvsg::ViewStateWriteLock(viewState)->setScene( scene.m_sceneHandle );

  // setup the view state with some default values, e.g. create a camera looking at the scene
  setupDefaultViewState(viewState);

  // SUNDOG: Create the SilverLining manager class and hook it into the scene.
  silverLining = new SilverLiningManager();
  if (silverLining)
  {
     NodeSharedPtr root = ReadableObject<Scene>(scene.m_sceneHandle)->getRootNode();

     if (root)
     {
         // Get the root group node
         if ( !isPtrTo<Group>( root ) ) return 0;
         GroupSharedPtr group = sharedPtr_cast<Group>( root );

         // Add a directional light source (managed via SilverLiningManager each frame)
         DirectedLightHandle *lightHdl = DirectedLightHandle::create();
         {
            WritableObject<DirectedLight> light(lightHdl);
            light->setAmbientColor(Vec3f(1.0f, 1.0f, 1.0f));
            light->setDiffuseColor(Vec3f(1.0f, 1.0f, 1.0f));
            light->setSpecularColor(Vec3f(0.0f, 0.0f, 0.0f));
         }
         WritableObject<Group>(group)->addLightSource(lightHdl);
         
         // Push out the far clip plane to encompass the clouds
         CameraSharedPtr cameraHdl;
         ViewStateReadLock viewStateLock( viewState );
         cameraHdl = viewStateLock->getCamera();
         FrustumCameraWriteLock camLock(cameraHdl);
         camLock->setNearDistance(1.0f);
         camLock->setFarDistance(50000.0f);

         // Hook in SilverLining
         SilverLining::Atmosphere *atm = silverLining->AttachToRoot(group, cameraHdl, lightHdl);

         // Add in our clouds
         setupAtmosphericConditions(atm);
     }
  }
  
  // create a render context
  nvgl::SmartRenderContextGL renderContextGL = nvgl::RenderContextGL::create( nvgl::RenderContextGL::Attach() );

  // create a render target in the render context
  rtgl = nvgl::RenderTargetGLFB::create ( renderContextGL ); 

  // configure the render target to clear the depth buffer only
  rtgl->setClearMask( GL_DEPTH_BUFFER_BIT );

  // create a SceneRenderer for OpenGL
  srgl2 = nvgl::SceneRendererGL2::create();

  // pass it the render target and the view state
  srgl2->setRenderTarget( rtgl );
  srgl2->setViewState( viewState );

  // configure GLUT to call the display() and reshape() functions on the appropriate events
  glutDisplayFunc( &display );
  glutReshapeFunc( &reshape );

  // enter the GLUT main loop, start event processing
  glutMainLoop();

  // reset the scene renderer and the target
  rtgl.reset();
  srgl2.reset();

  // Clean up SilverLining
  if (silverLining) delete silverLining;

  // finish the sample's main function
  return 0;
}

// the main function
// initialize SceniX and GLUT, create a window and start the sample
int main( int argc, char *argv[] )
{
  // initialize SceniX
  nvsg::nvsgInitialize();

  // initialize GLUT, set window size and display mode, create the main window
  glutInit( &argc, argv );
  glutSetOption( GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_CONTINUE_EXECUTION );
  glutInitWindowSize( 640, 480 );
  glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_ALPHA );
  glutCreateWindow( "SceniX minimal GLUT example" );

  // call the sample's main function
  int result = sampleMain();

  // terminate SceniX
  nvsg::nvsgTerminate();

  // return the result returned by the sample function
  return result;
}
